(* Blocks of tableaux. *)
open Softcore open Constraint open Looping open Purity open Schema open Satisfiability open Literal open Stats

let trace = false
let print_endline x = if trace then print_endline x else ()

module Block =
  struct
    (* Blocks are just the pair of a set of schemas and of a constraint + various precomputed information. *)
    type t = Schema.t list * Constraint.t * Literal.t list

    (* Constructors & Selectors *)
    let of_schs_and_cstr ~cstr ~schs = (schs,cstr,[])
    let schs_of = Tuple3.get1
    let cstr_of = Tuple3.get2
    let model_of = Tuple3.get3
    let is_empty (schs,_,_) = match schs with [] -> true | _ -> false

    (* Print stuff *)
    let to_string (schs,cstr,_) = (Schema.list_to_string schs) ^ "  " ^ (Constraint.to_string ~parenthesized:true cstr)
    let print bck = print_endline (to_string bck)
    let model_to_string ((schs,cstr,model) as bck) ~exclude = 
      (* this should be called only after the execution of regstab so schs either is empty or contains only literals *)
      let model = 
        if is_empty bck then model else 
          List.fold_right schs ~init:model ~f:(fun s acc -> match s with (`Lit l : Schema.t) -> l::acc | _ -> assert false) 
      in
      let filtered_model = List.filter ~f:(fun l -> not (List.mem (Literal.prop_of l) ~set:exclude)) model in
      Softcore.list_to_string ~to_string:(Literal.to_string <|| try_instantiate ~cstr) filtered_model

    (* Maximum upper bound of all schemas *)
    let max_bound bck = List.fold ~f:(fun current_max -> Schema.max_bound ~acc:current_max) ~init:None (schs_of bck)

    (* True if a block contains a schema containing iterations *)
    let contains_iterations bck = List.exists (schs_of bck) ~f:Schema.contains_iterations

    (* True if a block contains only literals *)
    let contains_only_literals bck = List.for_all (schs_of bck) ~f:(function (`Lit _:Schema.t) -> true | _ -> false)

    (* Purity interface *)
    type purable = t

    (* Just apply purity_add to each schema *)
    let purity_add bck ?(range=None) ~pur = List.fold (schs_of bck) ~f:(fun pur -> Schema.purity_add ~range:None ~pur) ~init:pur

    (* Removes all the pure literals of the set, adds them to the model *)
    let purify (schs,_,models) ~cstr ~pur =
      let m,non_pures,pures = List.fold schs ~init:(Modified.Unmodified,[],models) ~f:(fun acc sch ->
        match acc,Schema.purify ~cstr ~pur sch with
        |(m1,schs',models),(m2,(`Top:Schema.t)) -> 
            (match sch with
            |(`Lit l:Schema.t) -> Modified.max m1 m2,schs',l::models
            |`Top   -> Modified.max m1 m2,schs',models
            |_      -> assert false)
        |(m1,schs',models),(m2,sch)             -> Modified.max m1 m2,sch::schs',models)
      in
      m,(non_pures,cstr,pures)

    (* Computes the purity of the whole block *)
    let purity = purity_add ~range:None ~pur:Purity.empty

    (* Normalize the block.
     * Approx. definition:
     *   Let P_e or Op_i=f..e be the first elt of a list of schemata [S] s.t. e contains the parameter n.
     *   If there is no such e then [S] is already normalized.
     *   Else the normalized of [S] is the list which is just the same as [S] except that every occurrence is shifted by n-e.
     *   So after normalization e is equal to n.
     *)
    let normalize (schs,cstr,_) =
      let rec first_expression = function
        |[]         -> None
        |sch::schs' -> 
            (match Schema.expr_involving_param sch with
            |None   -> first_expression schs'
            |Some e -> Some e)
      in
      match first_expression schs with
      |None                  -> (schs,cstr,[])
      |Some first_expression -> begin
        let schs' =
          match Indexes.get_var first_expression with
          |None     -> assert false
          |Some var -> List.map schs ~f:(Schema.substitute ~var ~by:(Indexes.inverse first_expression))
        in
        let cstr' = Constraint.add_to_var cstr (-(Indexes.get_shift first_expression)) in
        (schs',cstr',[])
      end

    (* Comparison *)
    let rec schs_equal schs1 schs2 =
      match schs1,schs2 with
      |[],[]                     -> true
      |sch1::schs1',sch2::schs2' -> Schema.equal sch1 sch2 && schs_equal schs1' schs2'
      |_                         -> false
    let equal (schs1,cstr1,_) (schs2,cstr2,_) = Constraint.equal cstr1 cstr2 && schs_equal schs1 schs2

    (* Removal of duplicate schemata *)
    let dedup (schs,cstr,models) =
      let sorted_schs = List.sort ~cmp:(fun x y -> Schema.total_compare y x) schs in
      let rec dedup_schs ~acc = function
        |[]             -> acc
        |[sch]          -> sch::acc
        |sch1::sch2::schs -> if Schema.equal sch1 sch2 then dedup_schs (sch2::schs) ~acc else dedup_schs (sch2::schs) ~acc:(sch1::acc)
      in 
      (dedup_schs sorted_schs ~acc:[],cstr,models)

    (* Generic strategies to apply some given rules *)
    type branch_update =
      |Replace_with_schema                of Schema.t
      |Replace_with_schemas               of Schema.t * Schema.t
      |Replace_with_schema_add_constraint of Schema.t * Constraint.t
      |Add_constraint                     of Constraint.t

    type branching =
      |Go_on
      |Close
      |Update of branch_update
      |Branch of branch_update * branch_update

    let apply_rules_with_closure (schs,cstr,models) ~stat ~rules ~k =
      let rec traverse ~schs ~cstr ~models ~acc ~litmap =
        print_endline (to_string (of_schs_and_cstr ~schs ~cstr));
        match schs with
        |[]                                                        -> k (acc,cstr,models)
        |((`Lit l):Schema.t)::_ when ClosureSet.closure_applies litmap ~key:l -> stat incr.closed; print_endline "closed!"; NoModelFound
        |sch::schs'                                                -> 
            let update ~k = function
              |Replace_with_schema sch'                       -> k (traverse ~schs:(sch'::schs') ~cstr ~acc ~litmap ~models)
              |Replace_with_schemas(sch1,sch2)                -> k (traverse ~schs:(sch1::sch2::schs') ~cstr ~acc ~litmap ~models)
              |Replace_with_schema_add_constraint(sch',cstr') -> k (traverse ~schs:(sch'::schs') ~cstr:(bounds_conjunction cstr cstr') ~acc ~litmap ~models)
              |Add_constraint cstr'                           -> k (traverse ~schs:schs' ~cstr:(bounds_conjunction cstr cstr') ~acc ~litmap ~models)
            in
            (match rules ~sch ~cstr with
            |Go_on         -> traverse ~schs:schs' ~cstr ~models ~acc:(sch::acc) ~litmap:(
                match sch with ((`Lit l):Schema.t) -> ClosureSet.add litmap ~key:l | _ -> litmap)
            |Close         -> stat incr.closed; NoModelFound
            |Update u      -> update ~k:ident u
            |Branch(u1,u2) -> let second_branch = function ModelFound c -> ModelFound c | NoModelFound -> update ~k:ident u2 in update ~k:second_branch u1)
      in
      traverse ~schs ~cstr ~models ~acc:[] ~litmap:(ClosureSet.empty ~cstr)

    let apply_rules (schs,cstr,models) ~stat ~rules ~k =
      print_endline (to_string (of_schs_and_cstr ~schs ~cstr));
      let rec traverse ~schs ~cstr ~models ~acc =
        match schs with
        |[]                                                       -> k (acc,cstr,models)
        |sch::schs'                                               -> 
            let update ~k = function
              |Replace_with_schema sch'                       -> k (traverse ~schs:(sch'::schs') ~cstr ~acc ~models)
              |Replace_with_schemas(sch1,sch2)                -> k (traverse ~schs:(sch1::sch2::schs') ~cstr ~acc ~models)
              |Replace_with_schema_add_constraint(sch',cstr') -> k (traverse ~schs:(sch'::schs') ~cstr:(bounds_conjunction cstr cstr') ~acc ~models)
              |Add_constraint cstr'                           -> k (traverse ~schs:schs' ~cstr:(bounds_conjunction cstr cstr') ~acc ~models)
            in
            (match rules ~sch ~cstr with
            |Go_on         -> traverse ~schs:schs' ~cstr ~models ~acc:(sch::acc)
            |Close         -> stat incr.closed; NoModelFound
            |Update u      -> update ~k:ident u
            |Branch(u1,u2) ->
                let second_branch = function ModelFound c -> ModelFound c | NoModelFound -> update ~k:ident u2 in
                update ~k:second_branch u1)
      in
      traverse ~schs ~cstr ~models ~acc:[]
      
  end

include Block

open Looping 

(* Finally we can construct the trie from all the maps of other data strutures.
 *
 * Important note: blocks are not considered as sets but indeed as lists.
 * Termination property still holds and this is much more efficient (and readable) this way.
 *)
module LoopingMap =
  struct
    module SchMap = Schema.LoopingMap
    type loopable = Block.t
    type 'a t = ('a t) SchMap.t * (Constraint.t * 'a) list (* Trie structure *)
    (* TOTRY:
     * - integrate constraint? i.e. ('a * Constraint.t) option; I think it's useless *edit:NO!!!!*
     * - add a constructor Node of ... : allow to get rid of -rectypes; maybe
     * there's a loss in efficiency 
     *)
    type param = unit

    let empty = (SchMap.empty,[])
    
    let add m ~key ~data =
      let rec add' (schmap,current_data) ~schs ~data ~cstr =
        (match schs with
        |[]         -> schmap, ((cstr,data)::current_data)
        |sch::schs' ->
          (SchMap.add schmap ~key:sch ~data:(add' (Option.value (SchMap.find schmap ~key:sch) ~default:empty) ~schs:schs' ~data ~cstr)), current_data)
      in
      let (schs,cstr,_) = normalize key in
      add' m ~schs ~data ~cstr

    let find m ~key =
      let rec find' (m,data) ~schs ~cstr =
        match schs with
        |[]         -> begin try Some (List.assoc cstr data) with _ -> None end
        |sch::schs' ->
            (match SchMap.find m ~key:sch with
            |None    -> None
            |Some m' -> find' m' ~schs:schs' ~cstr)
      in
      let (schs,cstr,_) = normalize key in
      find' m ~schs ~cstr

    let rec fold_data (m,data) ~f ~init = 
      let init = SchMap.fold_data m ~init ~f:(fun ~data init -> fold_data data ~f ~init) in
      List.fold ~init ~f:(fun acc (_,data) -> f ~data acc) data

    let size = fold_data ~f:(fun ~data:_ -> succ) ~init:0
  end
