(* 
 * Arithmetic constraints.
 * See developper doc first.
 *)
open Softcore

(* Kind of equality *)
type inequality = Le | Gt | Eq

  (* Various facilities:
   * - conversion to string
   * - reversing an inequality (x > y is equivalent to y <= x-1 and *not to* y <= x, but x = y is equivalent to y = x,
   *   so reverse_sgn gives the reverse sign and reverse_int gives the function to apply to the integer in order to get the equivalence :
   *   fun x -> x-1 if inequality, fun x -> x if equality)
   * - conversion to the corresponding functions of ocaml
   *)
let ineq_to_string = function Le -> " <= " | Gt -> " > "  | Eq -> " = "
let reverse_sgn    = function Le -> Gt   | Gt -> Le   | Eq -> Eq
let reverse_int    = function Le         | Gt -> pred | Eq -> ident
let to_fun ineq (x:int) (y:int) = match ineq with Le -> x <= y | Gt -> x > y | Eq -> x = y 
let ineq_equal ineq1 ineq2 =
  match ineq1,ineq2 with
  |Le,Le | Gt,Gt | Eq,Eq -> true
  |_                     -> false

(* Type of the constraints, either true, or false, or the comparison between a variable and an int
 * i.e. the only non-trivial constraints are of the form n {<=,>,=} k where n is a variable, k an integer
 * As a consequence, n < 4 & n > 0 is not a supported constraint !
 * *We don't need more for regular schemata*, the strategy implies that such constraints should not occur.
 *)
type t =
  |TrueConstraint
  |FalseConstraint
  |AtomConstraint of Variable.t * inequality * int
  |BetweenConstraint of Variable.t * int * int

let equal c1 c2 =
  match c1,c2 with
  |TrueConstraint,TrueConstraint                           -> true
  |FalseConstraint,FalseConstraint                         -> true
  |AtomConstraint(v1,ineq1,n1),AtomConstraint(v2,ineq2,n2) -> n1 = n2 && ineq_equal ineq1 ineq2 && Variable.equal v1 v2
  |BetweenConstraint(v1,n1,p1),BetweenConstraint(v2,n2,p2) -> n1 = n2 && p1 = p2 && Variable.equal v1 v2
  |_                                                       -> false

(* Print stuff *)
let to_string ?(parenthesized=false) = 
  let paren s = if parenthesized then s else "" in
  function
    |TrueConstraint -> ""
    |FalseConstraint -> "(FALSE)"
    |AtomConstraint(v,ineq,n) -> paren "(" ^ Variable.to_string v ^ ineq_to_string ineq ^ string_of_int n ^ paren ")"
    |BetweenConstraint(v,n1,n2) -> paren "(" ^ string_of_int n1 ^ ineq_to_string Le ^ Variable.to_string v ^ ineq_to_string Le ^ string_of_int n2 ^ paren ")"
let print = print_endline <|| to_string

(* Exceptions *)
exception NonNormalizedConstraint
exception UnsupportedConstraint
exception NonRegularBounds

(* Selectors *)
let of_bool = function true -> TrueConstraint | false -> FalseConstraint
let is_true c = match c with TrueConstraint -> true | _ -> false
let is_false c = match c with FalseConstraint -> true | _ -> false
let is_trivial c = match c with FalseConstraint | TrueConstraint -> true | _ -> false

let normalized = function
  |BetweenConstraint(_,n1,n2) when n2 <= n1 -> raise NonNormalizedConstraint
  |_ as c -> c

let is_equality c = match normalized c with AtomConstraint(_,Eq,_) -> true | _ -> false
let is_lower = function AtomConstraint(_,Le,_) -> true | _ -> false
let number_of = function AtomConstraint(_,Eq,n) -> Some n | _ -> None

(* Constructors *)
let top     = TrueConstraint

let add_to_var c k =
  match c with
  |TrueConstraint | FalseConstraint -> c
  |AtomConstraint(v,ineq,n) -> AtomConstraint(v,ineq,n-k)
  |BetweenConstraint(v,n1,n2) -> BetweenConstraint(v,n1-k,n2-k)

(* Two possibilities to create a constraint, see developper doc.
 * Roughly: only one of two bounds can contain a variable.
 * If regular_bounds is true then we don't authorize each term to have a variable.
 *)
let constraint_atom_internal ~regular_bounds t1 ineq t2 =
  match (Indexes.get_var t1),(Indexes.get_var t2) with
  |Some(v),None                              -> AtomConstraint(v, ineq, (Indexes.get_shift t2)-(Indexes.get_shift t1))
  |None, Some(v)                             -> AtomConstraint(v,reverse_sgn ineq, reverse_int ineq ((Indexes.get_shift t1)-(Indexes.get_shift t2)))
  |None,None        when regular_bounds      -> raise NonRegularBounds
  |Some(v),Some(v') when regular_bounds      -> raise NonRegularBounds
  |Some(v),Some(v') when Variable.equal v v' -> of_bool (to_fun ineq (Indexes.get_shift t1) (Indexes.get_shift t2))
  |None,None                                 -> of_bool (to_fun ineq (Indexes.get_shift t1) (Indexes.get_shift t2))
  |_                                         -> raise UnsupportedConstraint

let constraint_atom = constraint_atom_internal ~regular_bounds:false
let bounds_constraint = constraint_atom_internal ~regular_bounds:true 

let ensure_has_variable c v =
  match c with
  | TrueConstraint | FalseConstraint -> c
  | AtomConstraint(v',_,_) | BetweenConstraint(v',_,_) when v'=v -> c
  | _ -> raise UnsupportedConstraint

let same_variables c1 c2 = 
  match c1 with
  | TrueConstraint | FalseConstraint -> c1,c2
  | AtomConstraint(v,_,_) | BetweenConstraint(v,_,_) -> c1, ensure_has_variable c2 v

(* Combines two constraints.
 * As previously said we don't support general conjunction of constraints.
 * So constraints can be combined only if one is trivial or if both deal with the same variable in some "convenient" way.
 *)
let bounds_conjunction c1 c2 =
  match same_variables (normalized c1) (normalized c2) with
  |TrueConstraint,c
  |c,TrueConstraint                                                              -> c
  |FalseConstraint,_
  |_,FalseConstraint                                                             -> FalseConstraint
  |AtomConstraint(v,Eq,n),AtomConstraint(_,ineq,n') when not (to_fun ineq n n') -> FalseConstraint
  |AtomConstraint(v,Eq,n),AtomConstraint(_,ineq,n')                             -> AtomConstraint(v,Eq,n)
  |AtomConstraint(v,Le,n),AtomConstraint(_,Le,n')                               -> AtomConstraint(v,Le,min n n')
  |AtomConstraint(v,Gt,n),AtomConstraint(_,Gt,n')                               -> AtomConstraint(v,Gt,max n n')
  |AtomConstraint(v,Le,n),AtomConstraint(_,Gt,n')   when n<=n'                  -> FalseConstraint
  |AtomConstraint(v,Gt,n),AtomConstraint(_,Le,n')   when n>=n'                  -> FalseConstraint
  |AtomConstraint(v,Le,n),AtomConstraint(_,Gt,n')   when n=n'+1                 -> AtomConstraint(v,Eq,n)
  |AtomConstraint(v,Gt,n),AtomConstraint(_,Le,n')   when n'=n+1                 -> AtomConstraint(v,Eq,n')
  |AtomConstraint(v,Gt,n),AtomConstraint(_,Le,n')                               -> BetweenConstraint(v,n+1,n')
  |AtomConstraint(v,Le,n),AtomConstraint(_,Gt,n')                               -> BetweenConstraint(v,n'+1,n)
  |BetweenConstraint(v,n1,n2),AtomConstraint(_,Le,n')                           -> BetweenConstraint(v,n1,min n2 n')
  |BetweenConstraint(v,n1,n2),AtomConstraint(_,Gt,n')                           -> BetweenConstraint(v,max n1 (n'+1),n2)
  |AtomConstraint(_,Le,n'),BetweenConstraint(v,n1,n2)                           -> BetweenConstraint(v,n1,min n2 n')
  |AtomConstraint(_,Gt,n'),BetweenConstraint(v,n1,n2)                           -> BetweenConstraint(v,max n1 (n'+1),n2)
  |BetweenConstraint(v,n1,n2),AtomConstraint(_,Eq,n') when n1 <= n' && n' <= n2 -> c2
  |AtomConstraint(_,Eq,n'),BetweenConstraint(v,n1,n2) when n1 <= n' && n' <= n2 -> c2
  |AtomConstraint(_,Eq,n'),BetweenConstraint(v,n1,n2)                           -> FalseConstraint
  |BetweenConstraint(v,n1,n2),AtomConstraint(_,Eq,n')                           -> FalseConstraint
  |_                                                                             -> raise UnsupportedConstraint

(* Consequence relation between those constraints. *)
let entails hyp con =
  match same_variables (normalized hyp) (normalized con) with
  |FalseConstraint,_
  |_,TrueConstraint                                  -> true
  |TrueConstraint,_
  |_,FalseConstraint                                 -> false
  |AtomConstraint(_,Eq,n),AtomConstraint(_,ineq,n')  -> to_fun ineq n n'
  |AtomConstraint(_,Le,n),AtomConstraint(_,Le,n')
  |BetweenConstraint(_,_,n),AtomConstraint(_,Le,n')  -> n <= n'
  |AtomConstraint(_,Gt,n),AtomConstraint(_,Gt,n')    -> n >= n'
  |BetweenConstraint(_,n,_),AtomConstraint(_,Gt,n')  -> n-1 >= n'
  |AtomConstraint(_,Eq,n),BetweenConstraint(_,n1,n2) -> n1 <= n && n <= n2
  |_                                                 -> false

let is_acceptable_as_input ?with_bounds = function
  |AtomConstraint(_,Eq,_) | TrueConstraint | FalseConstraint -> true
  |AtomConstraint(_,Le,_) | BetweenConstraint(_,_,_) -> false
  |AtomConstraint(_,Gt,n) ->
      match with_bounds with
      |None      -> true
      |Some(l,u) -> n >= (Indexes.get_shift l)-(Indexes.get_shift u)-2


(* Some predefined relations. *)
let always_greater_than   ?(hyp=TrueConstraint) t1 t2 = entails hyp (constraint_atom t1 Gt t2)
let always_lower_than ?(hyp=TrueConstraint) t1 t2 = always_greater_than ~hyp t2 t1
let always_greater_or_equal ?(hyp=TrueConstraint) t1 t2 = entails hyp (constraint_atom t2 Le t1)

let always_different ?(hyp=TrueConstraint) t1 t2 =
  match (constraint_atom t1 Eq t2) with
  |TrueConstraint         -> false
  |FalseConstraint        -> true
  |AtomConstraint(v,Eq,n) -> 
      (match ensure_has_variable (normalized hyp) v with
      |FalseConstraint            -> true
      |TrueConstraint             -> false
      |AtomConstraint(_,Le,n')    -> n'< n
      |AtomConstraint(_,Gt,n')    -> n'>=n
      |AtomConstraint(_,Eq,n')    -> n'<>n
      |BetweenConstraint(_,n1,n2) -> n < n1 || n >= n2)
  |_                      -> raise UnsupportedConstraint

(* Normalization for closure
 * Roughly: for closure what we want to know is if two indexes are *always equal* to each other.
 * This depends on the constraints.
 * But the only kind of constraint that can matter is an equality constraint.
 * In which case a trivial normal form can be obtained by just subtituting the parameter by its value in the constraint.
 *
 * Having a normal form allows to use classical data structures to test efficiently if a literal occurs in the structure.
 *)
let normalize_index cstr ~idx =
  match normalized cstr with
  |AtomConstraint(var,Eq,n) -> Indexes.substitute idx ~var ~by:(Indexes.of_num n)
  |_                        -> idx

let try_instantiate_index = normalize_index
