(** Arithmetic constraints.
 *
 * Constraints have the form t1 \{<=,>,=\} t2 where t1,t2 are index expressions.
 * @author Vincent Aravantinos
 *)

(** Kind of inequality.
 *
 *)
type inequality =
  |Le (** Lower or Equal *)
  |Gt (** Greater Than *)
  |Eq (** Equal *)

(** Type of constraints. *)
type t

exception NonNormalizedConstraint
exception UnsupportedConstraint
exception NonRegularBounds

(** {6 Constructors } *)

(** The constraint constantly true. *)
val top                : t

(** Shift the variable in the constraint by adding an integer to it. *)
val add_to_var         : t -> int -> t

(** A constraint containing just the comparison of two indexes. *)
val constraint_atom    : Indexes.t -> inequality -> Indexes.t -> t

(** A constraint generated from the comparison of two iteration bounds of regular schemata.
 * Having this knowledge allows to discard earlier some cases that should not occur. 
 * @raise NonRegularBounds if the parameters cannot come from iteration bounds.
 * *)
val bounds_constraint  : Indexes.t -> inequality -> Indexes.t -> t

(** The conjunction of two constraints. *)
val bounds_conjunction : t -> t -> t

(** {6 Selector }*)

val is_trivial : t -> bool
val is_true : t -> bool
val is_false : t -> bool
val is_equality : t -> bool
val is_lower : t -> bool
val number_of : t -> int option
val is_acceptable_as_input : ?with_bounds:(Indexes.t * Indexes.t) -> t -> bool

(** {6 Print stuff} *)

val to_string : ?parenthesized:bool -> t -> string
val print     : t -> unit

val equal : t -> t -> bool

(** Instantiate an index if the constraint is an equality *)
val try_instantiate_index  : t -> idx:Indexes.t -> Indexes.t

(** {6 Comparison of terms, taking constraints into account } *)

val always_greater_than     : ?hyp:t -> Indexes.t -> Indexes.t -> bool
val always_lower_than       : ?hyp:t -> Indexes.t -> Indexes.t -> bool
val always_greater_or_equal : ?hyp:t -> Indexes.t -> Indexes.t -> bool
val always_different        : ?hyp:t -> Indexes.t -> Indexes.t -> bool

(** {6 Normalization of a term for better detection of closure} *)
val normalize_index  : t -> idx:Indexes.t -> Indexes.t

