(** Literals.
 *
 * I.e. P_n, P_n+1, P_1, P_i, ~P_i, ~P_1, etc.
 *
 * @author Vincent Aravantinos
 *)

type t

(** {6 Constructors} *)

(** A negative literal (i.e. of the form ~P_e) *)
val neg : Proposition.t -> Indexes.t -> t
(** A positive literal (i.e. of the form P_e) *)
val pos : Proposition.t -> Indexes.t -> t

(** {6 Selectors} *)

(** Returns the pair [(prop,idx)] i.e. forget the sign of the literal *)
val atom_of : t -> Proposition.t * Indexes.t
val prop_of : t -> Proposition.t
val idx_of  : t -> Indexes.t
val is_pos  : t -> bool
val is_neg  : t -> bool

(** {6 Printing} *)

val to_string       : t -> string
val try_instantiate : t -> cstr:Constraint.t -> t

(** {6 Comparisons } *)
val compare : t -> t -> int
val equal   : t -> t -> bool

(** {6 Substitutions } *)

(** Standard substitution of parameters *)
val compute_lit  : t -> int -> t
include Substitution.Substituable with type substituable = t and type term = Indexes.t and type variable = Variable.t

(** "Propositional variable Substitution", a module to allow substitution of propositional variables by other propositional variables. *)
module PropSubs : Substitution.Substituable with type substituable = t and type term = Proposition.t and type variable = Proposition.t

(** {6 Looping } *)
module LoopingMap : Looping.LoopingMap with type loopable = t

(** {6 Closure } *)

(** To detect application of closure rule (rather) efficiently *)
module ClosureSet :
  sig
    type t
    val empty           : cstr:Constraint.t -> t
    val add             : t -> key:substituable -> t
    val closure_applies : t -> key:substituable -> bool
    val add_constraint  : t -> cstr:Constraint.t -> t
  end

(** {6 Others } *)

val opposite     : t -> t
