open Softcore open Lexing open Schema open Verbose

let non_regular_string       = "Error: the input schema is not regular, "
let more_than_one_string     = "more than one parameter was detected, bad parenthesing?"
let diff_lower_bounds_string = "two iterations with different lower bounds were detected."
let diff_upper_bounds_string = "two iterations with different upper bounds were detected."
let inside_iteration_string  = "the parameter occurs inside an iteration."

(* The main function, called with the input channel *)
let main ~verbose ?fname ch =
  if ch = stdin then print_endline "Taking input from stdin.";
  let lexbuf = 
    let lexbuf' = from_channel ch in
    match fname with None -> lexbuf' | Some fn -> { lexbuf' with lex_curr_p = { lexbuf'.lex_curr_p with pos_fname = fn } } in
  print_endline 
  (try
    let (sch,cstr) = Parser.main Lexer.token lexbuf in
    (if is_verbose verbose then "\nConjecture:\n  " ^ Schema.to_string sch ^ "\n\n" else "")
    ^
    (match Schema.is_regular sch with
    |`No argument -> non_regular_string ^ "\n" ^
      (match argument with
      |Different_lower_bounds -> diff_lower_bounds_string
      |Different_upper_bounds -> diff_upper_bounds_string
      |More_than_one_param    -> more_than_one_string
      |Param_inside_iteration -> inside_iteration_string
      )
    |`Yes bounds ->
        if not (Constraint.is_acceptable_as_input ?with_bounds:bounds cstr)
        then "Invalid constraint (the input constraint shall entail that the length of iterations is positive) - search manual for 'positive length'"
        else begin
          let result,verbose_funs = Regstab.regstab ~cstr ~verbose sch in
          (if is_verbose verbose && not (Constraint.is_true cstr) then "We consider " ^ Constraint.to_string cstr ^ "." else "")
          ^
          verbose_funs.stat Stats.to_string ^ verbose_funs.lemma Lemmas.Info_lemmas.to_string
          ^
          (match result with
          |Satisfiability.ModelFound bck -> "SATISFIABLE"
            ^ if verbose.print_model then 
              let cstr = Block.cstr_of bck in
              let empty_block_sentence = 
                if Block.is_empty bck then "" else
                  (" for any value of the parameter which is " ^ (if Constraint.is_lower cstr then "small" else "big")^ " enough")
              in
              begin
                "\n\n" ^
                begin
                  if Constraint.is_trivial cstr then ("Found model" ^ empty_block_sentence)
                  else match Constraint.number_of cstr with
                  |Some n ->
                    "for " ^ Constraint.to_string cstr ^ ", the corresponding instance is:\n\n" ^
                    Formula.to_string (Schema.compute_schema sch n) ^ "\n\nand the following is a model"
                  |None ->
                    "for any " ^ Constraint.to_string cstr ^ ", the corresponding instance is:\n" ^
                    Schema.to_string (Schema.algebraic_simplification sch ~cstr) ^ "\n\nand the following is a model" ^ empty_block_sentence
                end
                ^
                (match verbose.excluded_variables with
                |[] -> ""
                |xs -> " (variables " ^ Softcore.list_to_string ~to_string:Proposition.to_string ~sep:", " ~delim:("","") xs ^ " are excluded)")
                ^
                ":\n\n"
                ^ Block.model_to_string ~exclude:verbose.excluded_variables bck 
              end
            else ""
          |Satisfiability.NoModelFound   -> "UNSATISFIABLE")
        end)
  with Failure s -> s
  );
  print_newline ();
  flush stdout

(* Handling of command-line options, calls main *)
let _ =
  let channel = ref None in
  let verbose = ref false in
  let print_lemmas = ref false in
  let print_model = ref true in
  let excluded_vars = ref [] in
  let parse_excluded_vars vs = excluded_vars := List.map ~f:Proposition.of_string (String.split ~on:',' vs) in
  let options = [
    "--exclude-vars v1,v2,...", Arg.String(parse_excluded_vars), " If the schema is satisfiable, exclude the given variables of the printed model";
    "-l", Arg.Set(print_lemmas), " Print lemmas as they are inserted in the database";
    "-M", Arg.Clear(print_model), " Do not print a model when the schema is satisfiable";
    "--no-model", Arg.Clear(print_model), " Same as -M";
    "--print-lemmas", Arg.Set(print_lemmas), " Same as -l";
    "--verbose", Arg.Set(verbose), " Be verbose";
    "-v", Arg.Set(verbose), " Same as --verbose";
    "-x", Arg.String(parse_excluded_vars), " Same as --exclude-vars";
  ] in
  let options = Arg.align options in
  let usage = "regstab [file] - says if the schema in file is satisfiable, schema taken on stdin if no file provided" in
  let file s = 
    match !channel with
    |None   -> channel := Some(s)
    |Some _ -> failwith "Too many arguments."
  in
  Arg.parse options file usage;
  let verbose = Verbose.gather !verbose !print_lemmas !print_model !excluded_vars in
  match !channel with
  |None      -> main stdin ~verbose
  |Some file -> read_wrap ~f:(main ~verbose ~fname:file) file

